/*
 * $Id$
 *
 * This file is part of the MIPL Mobile IPv6 for Linux.
 *
 * Author: Henrik Petander <petander@tcs.hut.fi>
 *
 * Copyright 2003-2005 GO-Core Project
 *
 * MIPL Mobile IPv6 for Linux is free software; you can redistribute
 * it and/or modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; version 2 of
 * the License.
 *
 * MIPL Mobile IPv6 for Linux is distributed in the hope that it will
 * be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIPL Mobile IPv6 for Linux; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 * 02111-1307 USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#ifdef HAVE_LIBPTHREAD
#include <pthread.h>
#else
#error "POSIX Thread Library required!"
#endif
#include <errno.h>

#include "debug.h"
#include "bcache.h"
#include "xfrm.h"
#include "util.h"
#include "tunnelctl.h"
#include "keygen.h"
#include "mh.h"
#include "cn.h"
#include "vt.h"

#define BCACHE_BUCKETS 32

#define BC_DEBUG_LEVEL 1

#if BC_DEBUG_LEVEL >= 1
#define BDBG dbg
#else
#define BDBG(...)
#endif

static struct hash bc_hash;

static int bcache_count = 0;

pthread_rwlock_t bc_lock; /* Protects binding cache */

/** 
 * get_bcache_count - returns number of home and cache entries
 * @type: HOMEREGENTRY, CACHEENTRY or ANY
 **/
int get_bcache_count(int type)
{
	if (!type)
		return bcache_count;
	return 0;
}

void dump_bce(struct bcentry *bce)
{
	switch(bce->type) {
	case BCE_NONCE_BLOCK:
		BDBG("NONCE_BLOCK entry: \n");
		break;
	case BCE_CACHED:
		BDBG("CACHED entry: \n");
		break;
	case BCE_HOMEREG:
		BDBG("HOMEREG entry: \n");
		break;
	default:
		BDBG("Unknown BCE type. \n");
	}
	BDBG("coa = %x:%x:%x:%x:%x:%x:%x:%x\n", NIP6ADDR(&bce->coa));
	BDBG("hoa = %x:%x:%x:%x:%x:%x:%x:%x\n", NIP6ADDR(&bce->peer_addr));
	BDBG("Local address = %x:%x:%x:%x:%x:%x:%x:%x\n",
	     NIP6ADDR(&bce->our_addr));
	BDBG(" lifetime = %d\n ", bce->lifetime.tv_sec);
	BDBG("seqno. %d\n", bce->seqno);
}

/**
 * _expire - expire binding cache entry
 **/
static void _expire(struct tq_elem *tqe)
{	
	pthread_rwlock_wrlock(&bc_lock);
	if (!task_interrupted()) {
		struct bcentry *e = tq_data(tqe, struct bcentry, tqe);
	
		pthread_rwlock_wrlock(&e->lock);

		if (e->type == BCE_CACHED) {
			/* To do: MN needs to reverse tunnel this */
			mh_send_brr(&e->peer_addr, &e->our_addr);
			e->type = BCE_CACHE_DYING;
			add_task_rel(&CN_BRR_BEFORE_EXPIRY_TS,
				     &e->tqe, _expire);
			pthread_rwlock_unlock(&e->lock);
			pthread_rwlock_unlock(&bc_lock);
			return;
		}
		if (e->type != BCE_NONCE_BLOCK) {
			xfrm_del_bce(&e->our_addr, &e->peer_addr);
			bcache_count--;
		}
		if (e->type == BCE_CACHE_DYING) {
			struct timespec minlft;

			BDBG("Changing BC_CACHE entry to nonce block entry\n");
			
			if (!rr_cn_nonce_lft(e->nonce_hoa, &minlft)) {
				clock_gettime(CLOCK_REALTIME, &e->add_time);
				e->type = BCE_NONCE_BLOCK;
				tssub(minlft, e->add_time, e->lifetime);
				add_task_abs(&minlft, &e->tqe, _expire);
				pthread_rwlock_unlock(&e->lock);
				pthread_rwlock_unlock(&bc_lock);
				return;
			}
		}
		hash_delete(&bc_hash, &e->our_addr, &e->peer_addr);

		if (e->cleanup)
			e->cleanup(e);
		BDBG("deleting the entry: \n");
		dump_bce(e);
		pthread_rwlock_unlock(&e->lock);
		bcache_free(e);
	}
	pthread_rwlock_unlock(&bc_lock);
}
	
/**
 * bcache_alloc - allocate binding cache entry
 * @type: type of entry
 *
 * Allocates a new binding cache entry.  @type may be %BCE_HOMEREG or
 * %BCE_CACHED.  Returns allocated space for an entry or NULL if none
 * available.
 **/
struct bcentry *bcache_alloc(int type)
{
	struct bcentry *tmp;
	/* This function should really return space from a
	 * pre-allocated pool.  If pool is empty, but type is
	 * BCE_HOMEREG, already existing BCE_CACHED entry may be
	 * replaced. */
	tmp = malloc(sizeof(struct bcentry));

	if (tmp == NULL)
		return NULL;

	if (pthread_rwlock_init(&tmp->lock, NULL)) {
		free(tmp);
		return NULL;
	}
	memset(tmp, 0, sizeof(*tmp));
	INIT_LIST_HEAD(&tmp->tqe.list);
	return tmp;
}

/**
 * bcache_free - release allocated memory
 * @bce: BC entry to free
 *
 * Release allocated memory back to unused pool.
 **/
void bcache_free(struct bcentry *bce)
{
	/* This function should really return allocated space to free
	 * pool. */
	pthread_rwlock_destroy(&bce->lock);
	free(bce);
}

/**
 * bul_get - returns a binding update list entry
 * @hinfo: home address info, optional if our_addr is present
 * @our_addr: local address (home address)
 * @peer_addr: address of CN
 *
 * Returns reference to non-null entry on success and null on failure.
 * If caller adjusts lifetime of entry, caller must call
 * bcache_update_expire() for BC entry to reschedule.  Caller must
 * call bcache_release_entry() after entry is not used anymore.
 **/
struct bcentry *bcache_get(const struct in6_addr *our_addr,
			   const struct in6_addr *peer_addr)
{
	struct bcentry *bce;

	assert(peer_addr && our_addr);

	pthread_rwlock_rdlock(&bc_lock);

	bce = hash_get(&bc_hash, our_addr, peer_addr);

	if (bce) 
		pthread_rwlock_wrlock(&bce->lock);
	else
		pthread_rwlock_unlock(&bc_lock);

	return bce;
}

/**
 * bcache_release_entry - unlocks a binding cache entry 
 **/
void bcache_release_entry(struct bcentry *bce)
{
	assert(bce);
	pthread_rwlock_unlock(&bce->lock);
	pthread_rwlock_unlock(&bc_lock);
}

/**
 * bce_type - get type of binding cache entry
 * @our_addr: our IPv6 address
 * @peer_addr: peer's IPv6 address
 *
 * Looks up entry from binding cache and returns its type.  If not
 * found, returns -%ENOENT.
 **/
int bce_type(const struct in6_addr *our_addr, const struct in6_addr *peer_addr)
{
        struct bcentry *bce;
        int type;

        bce = bcache_get(our_addr, peer_addr);

        if (bce == NULL)
                return -ENOENT;

        type = bce->type;
        bcache_release_entry(bce);

        return type;
}

static int __bcache_insert(struct bcentry *bce)
{
	int ret;

	ret = hash_add(&bc_hash, bce, &bce->our_addr, &bce->peer_addr);
	if (ret)
		return ret;

	bcache_count++;
	return 0;
}

static int __bcache_start(struct bcentry *bce)
{
	struct timespec expires, tmp;

	tssub(bce->lifetime, CN_BRR_BEFORE_EXPIRY_TS, tmp);
	clock_gettime(CLOCK_REALTIME, &bce->add_time);
	tsadd(bce->add_time,
	      bce->type == BCE_HOMEREG ? bce->lifetime : tmp,
	      expires);
	add_task_abs(&expires, &bce->tqe, _expire); 
	xfrm_add_bce(&bce->our_addr, &bce->peer_addr, &bce->coa, 0);
	return 0;
}

/* Adds bce entry and adds a timer for expiry / resend.  Caller must
   fill all non-private fields of bce */
int bcache_add(struct bcentry *bce)
{
	int ret = 0;

	assert(bce);

	bce->unreach = 0;
	pthread_rwlock_wrlock(&bc_lock);

	if ((ret = __bcache_insert(bce)) != 0) {
		pthread_rwlock_unlock(&bc_lock);
		return ret;
	}
	__bcache_start(bce);

	pthread_rwlock_unlock(&bc_lock);

	return 0;
}

int bcache_add_homereg(struct bcentry *bce)
{
	int ret = 0;

	assert(bce);
	assert(bce->type == BCE_DAD);

	bce->unreach = 0;
	pthread_rwlock_wrlock(&bc_lock);

	ret = __bcache_insert(bce);

	pthread_rwlock_unlock(&bc_lock);

	return ret;
}

int bcache_complete_homereg(struct bcentry *bce)
{
	assert(bce);

	/* XXX: This function has no lock for BCE because
	 * bcache_update_expire() doesn't. Why?
	 */
	__bcache_start(bce);

	return 0;
}

int bcache_update_expire(struct bcentry *bce)
{
	struct timespec expires;

	clock_gettime(CLOCK_REALTIME, &bce->add_time);
	if (bce->type == BCE_HOMEREG)
		expires = bce->lifetime;
	else {
		bce->type = BCE_CACHED;
		tssub(bce->lifetime, CN_BRR_BEFORE_EXPIRY_TS, expires);
	}
	tsadd(expires, bce->add_time, expires);
	add_task_abs(&expires, &bce->tqe, _expire);	
	xfrm_add_bce(&bce->our_addr, &bce->peer_addr, &bce->coa, 1);

	return 0;
}

/**
 * bcache_delete - deletes a bul entry
 **/
void bcache_delete(const struct in6_addr *our_addr,
		   const struct in6_addr *peer_addr)
{
	struct bcentry *bce;

	pthread_rwlock_wrlock(&bc_lock);
	bce = hash_get(&bc_hash, our_addr, peer_addr);
	if (bce == NULL) {
		pthread_rwlock_unlock(&bc_lock);
		return;
	}
	pthread_rwlock_wrlock(&bce->lock);
	if (bce->type != BCE_DAD) {
		del_task(&bce->tqe);
		if (bce->type != BCE_NONCE_BLOCK)
			xfrm_del_bce(&bce->our_addr, &bce->peer_addr);
	}
	bcache_count--;
	if (bce->cleanup)
		bce->cleanup(bce);
	if (bce->type == BCE_CACHED || bce->type == BCE_CACHE_DYING) {
		struct timespec minlft;
		if (!rr_cn_nonce_lft(bce->nonce_hoa, &minlft)) {
			bce->type = BCE_NONCE_BLOCK;
			add_task_abs(&minlft, &bce->tqe, _expire);
			pthread_rwlock_unlock(&bce->lock);
			pthread_rwlock_unlock(&bc_lock);
			return;
		}
	} 
	hash_delete(&bc_hash, &bce->our_addr, &bce->peer_addr);
	pthread_rwlock_unlock(&bce->lock);
	bcache_free(bce);
	pthread_rwlock_unlock(&bc_lock);
}

/**
 * bcache_init - initializes binding cache
 **/
int bcache_init(void)
{
	int ret;

	if (pthread_rwlock_init(&bc_lock, NULL))
		return -1;

	pthread_rwlock_wrlock(&bc_lock);
	ret = hash_init(&bc_hash, DOUBLE_ADDR, BCACHE_BUCKETS);
	pthread_rwlock_unlock(&bc_lock);

#ifdef ENABLE_VT
	if (ret < 0)
		return ret;

	ret = vt_bc_init();
#endif

	return ret;
}

/**
 * bcache_cleanup - cleans up a bcentry
 **/
static int bce_cleanup(void *data, void *arg)
{
	struct bcentry *bce = (struct bcentry *)data;
	if (bce->cleanup)
		bce->cleanup(bce);
	if (bce->type != BCE_NONCE_BLOCK)
		xfrm_del_bce(&bce->our_addr, &bce->peer_addr);
	if (bce->type != BCE_DAD)
		del_task(&bce->tqe);
	free(bce);

	return 0;
}

void bcache_flush(void)
{
	pthread_rwlock_wrlock(&bc_lock); 
	hash_iterate(&bc_hash, bce_cleanup, NULL);
	pthread_rwlock_unlock(&bc_lock);
}

void bcache_cleanup(void)
{
	pthread_rwlock_wrlock(&bc_lock); 
	hash_cleanup(&bc_hash);
	pthread_rwlock_unlock(&bc_lock);
}
/**
 * bcache_iterate - apply function to every BC entry
 * @func: function to apply
 * @arg: extra data for @func
 *
 * Iterates through binding cache, calling @func for each entry.
 * Extra data may be passed to @func in @arg.  @func takes a bcentry
 * as its first argument and @arg as second argument.
 **/
int bcache_iterate(int (* func)(void *, void *), void *arg)
{
	int err;
	pthread_rwlock_rdlock(&bc_lock); 
	err = hash_iterate(&bc_hash, func, arg);
	pthread_rwlock_unlock(&bc_lock);
	return err;
}
