/*
 * $Id$
 *
 * This file is part of the MIPL Mobile IPv6 for Linux.
 * 
 * Author: Antti Tuominen <anttit@tcs.hut.fi>
 *
 * Copyright 2003-2004 GO-Core Project
 *
 * MIPL Mobile IPv6 for Linux is free software; you can redistribute
 * it and/or modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; version 2 of
 * the License.
 *
 * MIPL Mobile IPv6 for Linux is distributed in the hope that it will
 * be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIPL Mobile IPv6 for Linux; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 * 02111-1307 USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <sys/uio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <pthread.h>
#include <netinet/in.h>
#include <syslog.h>

#include "debug.h"
#include "defpath.h"
#include "libmip6.h"
#include "bcache.h"
#include "util.h"

/* This section defines IPC messages used with libmip6.  These
 * messages MUST be sync'ed with values in libmip6/ipc.h.  They are
 * here (and not in a header file), since they should never be used
 * outside this file. */
#define MIP_GET_ONE_MN_INFO	0x0001
#define MIP_GET_ALL_MN_NUM	0x0002
#define MIP_GET_ALL_MN_INFO	0x0003
#define MIP_CATCH_EVENTS	0x0004
#define MIP_IGNORE_EVENTS	0x0005

#define MIP_ONE_MN_INFO		0x0011
#define MIP_ALL_MN_NUM		0x0012
#define MIP_ALL_MN_INFO		0x0013
#define MIP_EVENT		0x0014

#define MIP_LOCAL_INFO		0x0100

#define HI(w) w << 16
/* End of libmip6/ipc.h section */

static int events_enabled;
static int msqid_in, msqid_out;
static pthread_t mqh;

static int mninfo_lookup(struct in6_addr *home_addr, mobile_node_t *t)
{
	ipv6_addr_copy(&t->home_addr, home_addr);
	ipv6_addr_copy(&t->co_addr, home_addr);
	return 0;
}

/* Return total number of home addresses configured. */
static int local_mn_count(void)
{
	return 2;
}

/* Return all home addresses, with associated CoA and HA address */
static void local_mn_info(struct iovec **iov, int *count)
{
	mobile_node_t *t;

	iov = malloc(sizeof(struct iovec) * 2);
	*count = 2;
	iov[0]->iov_base = malloc(sizeof(mobile_node_t));
	iov[1]->iov_base = malloc(sizeof(mobile_node_t));
	t = (mobile_node_t *)iov[0]->iov_base;
}

/* Return total number of MNs in BC. */
static int bc_mn_count(void)
{
	return get_bcache_count(0);
}

struct bc_itargs {
	int index;
	int count;
	struct iovec *iovs;
	mobile_node_t *mn;
};

static int bc_iterator(void *v_bce, void *arg)
{

	struct bc_itargs *bc_a = (struct bc_itargs *)arg;
	struct bcentry *bce = (struct bcentry *)v_bce;
	if (bc_a->index < bc_a->count) {
		ipv6_addr_copy(&bc_a->mn[bc_a->index].home_addr, 
			       &bce->peer_addr);
		ipv6_addr_copy(&bc_a->mn[bc_a->index].co_addr, 
			       &bce->coa);
		ipv6_addr_copy(&bc_a->mn[bc_a->index].ha_addr, 
			       &bce->our_addr);
		bc_a->iovs[bc_a->index].iov_base = &bc_a->mn[bc_a->index]; 
		bc_a->iovs[bc_a->index].iov_len = sizeof(mobile_node_t);
		bc_a->index++;
		return 0;
	} else 
		return 1;
}
/* Return all MNs in Binding Cache. */
static void bc_mn_info(struct iovec **iov, int *count)
{
	struct bc_itargs args;
	args.count = bc_mn_count();
	args.iovs = malloc(args.count * sizeof(struct iovec));
	args.mn = malloc(args.count * sizeof(mobile_node_t));
	if (args.count && (!args.iovs || !args.mn))
		args.count = 0;
	args.index = 0;

	if (args.count)
		bcache_iterate(bc_iterator, &args);
	*count = args.count;
	*iov = args.iovs;
}

static void mq_mipapi_events(int enable)
{
	if (enable == 0)
		events_enabled = 0;
	else
		events_enabled = 1;
}

static int msqsnd_iovec(int mqid, int msgflg, long int mtype,
			const struct iovec *iov, size_t iov_count)
{
	size_t c, iovlen = 0;
	char *buffer, *p;
	int ret;
	
	for (c = 0; c < iov_count; c++) {
		iovlen += iov[c].iov_len;
	}

	buffer = malloc(sizeof(long int) + iovlen);
	if (buffer == NULL)
		return -ENOMEM;
	memset(buffer, 0, sizeof(*buffer));

	p = buffer;
	memcpy(p, &mtype, sizeof(long int));
	p += sizeof(long int);

	for (c = 0; c < iov_count; c++) {
		memcpy(p, iov[c].iov_base, iov[c].iov_len);
		p += iov[c].iov_len;
	}

	ret = msgsnd(mqid, (void *)buffer, iovlen, msgflg);
	free(buffer);

	return ret;
}

static void *mq_service(void *arg)
{
	while (1) {
		struct {
			long int mtype;
			pid_t pid;
			struct in6_addr home_addr;
		} msg;
		struct {
			long int mtype;
			mobile_node_t mn;
		} rep_1;
		struct {
			long int mtype;
			unsigned int count;
		} rep_n;

		int rcvd, count;
		struct iovec *iov;

		rcvd = msgrcv(msqid_in, &msg, sizeof(msg), 0, MSG_NOERROR);

		if (rcvd < sizeof(long int))
			continue;

		switch (msg.mtype) {
		case MIP_GET_ONE_MN_INFO:
			memset(&rep_1, 0, sizeof(rep_1));
			rep_1.mtype = HI(msg.pid) | MIP_ONE_MN_INFO;
			mninfo_lookup(&msg.home_addr, &rep_1.mn);
			msgsnd(msqid_out, &rep_1, sizeof(mobile_node_t), 0);
			break;
		case MIP_GET_ALL_MN_NUM:
			rep_n.mtype = HI(msg.pid) | MIP_ALL_MN_NUM;
			rep_n.count = bc_mn_count();
			msgsnd(msqid_out, &rep_n, sizeof(int), 0);
			break;
		case MIP_GET_ALL_MN_INFO:
			bc_mn_info(&iov, &count);
			dbg("MIP_GET_ALL_MN_INFO count: %d\n", count);
			msqsnd_iovec(msqid_out, 0,
				     HI(msg.pid) | MIP_ALL_MN_INFO,
				     iov, count);
			break;
		case MIP_GET_ALL_MN_NUM | MIP_LOCAL_INFO:
			rep_n.mtype = HI(msg.pid) | 
				MIP_ALL_MN_NUM | MIP_LOCAL_INFO;
			rep_n.count = local_mn_count();
			msgsnd(msqid_out, &rep_n, sizeof(int), 0);
			break;
		case MIP_GET_ALL_MN_INFO | MIP_LOCAL_INFO:
			local_mn_info(&iov, &count);
			break;
		case MIP_CATCH_EVENTS:
			mq_mipapi_events(1);
			break;
		case MIP_IGNORE_EVENTS:
			mq_mipapi_events(0);
			break;
		}
	}

	return NULL;
}

/* Call this when an event happens.  See Mobile IP API draft for
 * defined events. */
int mipapi_event(unsigned int event, const struct in6_addr *home_addr,
		 const struct in6_addr *coa, const struct in6_addr *ha_addr)
{
	struct msg_event {
		long int mtype;
		unsigned int event;
		mobile_node_t mn;
	} ev;

	if (!events_enabled) return 0;

	ev.mtype = MIP_EVENT;
	ev.event = event;
	ipv6_addr_copy(&ev.mn.home_addr, home_addr);
	ipv6_addr_copy(&ev.mn.co_addr, coa);
	ipv6_addr_copy(&ev.mn.ha_addr, ha_addr);
	msgsnd(msqid_out, &ev, sizeof(int) + sizeof(mobile_node_t), 0);

	return 0;
}

/* Call this to initialize message queues. */
int libmip6_ipc_init(void)
{
	key_t key_in, key_out;

        key_in = ftok(DEFAULT_PROGRAM_PATH, 0);
        key_out = ftok(DEFAULT_PROGRAM_PATH, 1);

	if (key_in < 0 || key_out < 0) {
		syslog(LOG_ERR, "libmip6_ipc_init failed\n");
		return -1;
	}
	/* enable IPC_EXCL if we can be sure message queues do not
	 * exist from previous mip6d instance */
        msqid_in = msgget(key_in, IPC_CREAT | /* IPC_EXCL | */
			  S_IRUSR | S_IWUSR | 
			  S_IRGRP | S_IWGRP | 
			  S_IROTH | S_IWOTH);

        msqid_out = msgget(key_out, IPC_CREAT | /* IPC_EXCL | */
			   S_IRUSR | S_IWUSR | 
			   S_IRGRP | S_IWGRP | 
			   S_IROTH | S_IWOTH);

	if (msqid_in < 0  || msqid_out < 0) {
		syslog(LOG_ERR, "libmip6_ipc_init: message queue init failed\n");
		return -1;
	}
	pthread_create(&mqh, NULL, mq_service, NULL);

	return 0;
}

/* Be sure to call this before exiting.  Otherwise, message queues are
 * left in the kernel. */
void libmip6_ipc_fini(void)
{
	mq_mipapi_events(0);
	pthread_cancel(mqh);
	pthread_join(mqh, NULL);
	msgctl(msqid_in, IPC_RMID, NULL);
	msgctl(msqid_out, IPC_RMID, NULL);
}
